//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "mecaSimu.h"

#include "TransProg/path.h"

AppliData ad;

real32_t
computeAcceleration(const real32_t *others,
                    const real32_t *pos,
                    const real32_t *speed,
                    int x,
                    int y,
                    int z,
                    real32_t accel[3])
{
real32_t maxPos=0.0f;
real32_t maxNeg=0.0f;
accel[0]=0.0f;
accel[1]=0.0f;
accel[2]=0.0f;
if(x&&(x!=NX-1))
  {
  for(int k=-1;k<=1;++k)
    {
    int oz=z+k;
    if((oz>=0)&&(oz<NZ))
      {
      for(int j=-1;j<=1;++j)
        {
        int oy=y+j;
        if((oy>=0)&&(oy<NY))
          {
          for(int i=-1;i<=1;++i)
            {
            int ox=x+i;
            if((ox>=0)&&(ox<NX))
              {
              int kind=(i?1:0)+(j?1:0)+(k?1:0);
              if(kind)
                {
                real32_t l0=ad.delta[kind-1];
                int id=MAKE_ID(ox,oy,oz);
                real32_t dx=others[id+OX]-pos[0];
                real32_t dy=others[id+OY]-pos[1];
                real32_t dz=others[id+OZ]-pos[2];
                real32_t f=sqrtf(dx*dx+dy*dy+dz*dz)*l0-1.0f;
                accel[0]+=f*dx;
                accel[1]+=f*dy;
                accel[2]+=f*dz;
                if(f>maxPos) maxPos=f;
                else if(f<maxNeg) maxNeg=f;
                }
              }
            }
          }
        }
      }
    }
  accel[0]*=K_ELAST/MASS;
  accel[1]*=K_ELAST/MASS;
  accel[2]*=K_ELAST/MASS;
  if((abs(x-ad.anchorX)<=1)&&
     (abs(y-ad.anchorY)<=1)&&
     (abs(z-ad.anchorZ)<=1))
    {
    accel[0]+=ad.forceX/(MASS*27.0f);
    accel[1]+=ad.forceY/(MASS*27.0f);
    accel[2]+=ad.forceZ/(MASS*27.0f);
    }
  accel[0]-=speed[0]*DAMP;
  accel[1]-=speed[1]*DAMP;
  accel[2]-=speed[2]*DAMP;
  accel[2]-=9.81f;
  }
return K_ELAST*(maxPos>-maxNeg ? maxPos : maxNeg);
}

void
prepareRendering(real32_t *spheres,
                 real32_t *colors,
                 real32_t *lines,
                 real32_t pos[3],
                 real32_t state,
                 int x,
                 int y,
                 int z,
                 int id)
{
int anchorDx=x-ad.anchorX;
int anchorDy=y-ad.anchorY;
int anchorDz=z-ad.anchorZ;
if((abs(anchorDx)<=1)&&
   (abs(anchorDy)<=1)&&
   (abs(anchorDz)<=1))
  {
  ++anchorDx;
  ++anchorDy;
  ++anchorDz;
  lines+=6*(anchorDx+3*(anchorDy+3*anchorDz));
  lines[0]=pos[0];
  lines[1]=pos[1];
  lines[2]=pos[2];
  lines[3]=pos[0]+0.0001f*ad.forceX;
  lines[4]=pos[1]+0.0001f*ad.forceY;
  lines[5]=pos[2]+0.0001f*ad.forceZ;
  }
#if defined GLOBAL_SPLIT_XYZ
  id*=3;
#endif
spheres+=id;
spheres[0]=pos[0];
spheres[1]=pos[1];
spheres[2]=pos[2];
colors+=id;
state/=1000.0f;
if(state>=0.0f)
  {
  state=0.2f+0.6f*(state<1.0f ? 1.0f-state : 0.0f);
  colors[0]=0.8f;
  colors[1]=state;
  colors[2]=state;
  }
else
  {
  state=0.2f+0.6f*(state>-1.0f ? 1.0f+state : 0.0f);
  colors[0]=state;
  colors[1]=state;
  colors[2]=0.8f;
  }
}

void
computePoint(int x,
             int y,
             int z,
             real32_t *spheres,
             real32_t *colors,
             real32_t *lines)
{
const real32_t *pIn=(real32_t *)0;
real32_t *sIn=(real32_t *)0;
real32_t *pOut=(real32_t *)0;
real32_t *sOut=(real32_t *)0;

#if defined USE_RK4
  real32_t dtOut=0.0;
  real32_t dtAcc=0.0;
  switch(ad.rk)
    {
    case 0:
      {
      pIn=ad.p0; sIn=ad.s0;
      pOut=ad.p2; sOut=ad.s2;
      dtOut=0.5f*DT;
      dtAcc=DT/6.0f;
      break;
      }
    case 1:
      {
      pIn=ad.p2; sIn=ad.s2;
      pOut=ad.p1; sOut=ad.s1;
      dtOut=0.5f*DT;
      dtAcc=DT/3.0f;
      break;
      }
    case 2:
      {
      pIn=ad.p1; sIn=ad.s1;
      pOut=ad.p2; sOut=ad.s2;
      dtOut=DT;
      dtAcc=DT/3.0f;
      break;
      }
    case 3:
      {
      pIn=ad.p2; sIn=ad.s2;
      pOut=ad.p1; sOut=ad.s1;
      dtOut=0.0f;
      dtAcc=DT/6.0f;
      break;
      }
    }
#else
  pIn=ad.p0; sIn=ad.s0; pOut=ad.p1; sOut=ad.s0;
#endif

int id=MAKE_ID(x,y,z);

sIn+=id;
pOut+=id;
sOut+=id;
real32_t pos[3]={ pIn[id+OX], pIn[id+OY], pIn[id+OZ] };
real32_t speed[3]={ sIn[OX], sIn[OY], sIn[OZ] };
real32_t accel[3];

real32_t state=computeAcceleration(pIn,pos,speed,x,y,z,accel);

#if defined USE_RK4
  real32_t *p3=ad.p3+id;
  real32_t *s3=ad.s3+id;
  switch(ad.rk)
    {
    case 0:
      {
      p3[OX]=pos[0]+dtAcc*speed[0];
      p3[OY]=pos[1]+dtAcc*speed[1];
      p3[OZ]=pos[2]+dtAcc*speed[2];
      s3[OX]=speed[0]+dtAcc*accel[0];
      s3[OY]=speed[1]+dtAcc*accel[1];
      s3[OZ]=speed[2]+dtAcc*accel[2];
      pos[0]+=dtOut*speed[0];
      pos[1]+=dtOut*speed[1];
      pos[2]+=dtOut*speed[2];
      speed[0]+=dtOut*accel[0];
      speed[1]+=dtOut*accel[1];
      speed[2]+=dtOut*accel[2];
      break;
      }
    case 1:
    case 2:
      {
      p3[OX]+=dtAcc*speed[0];
      p3[OY]+=dtAcc*speed[1];
      p3[OZ]+=dtAcc*speed[2];
      s3[OX]+=dtAcc*accel[0];
      s3[OY]+=dtAcc*accel[1];
      s3[OZ]+=dtAcc*accel[2];
      real32_t *p0=ad.p0+id;
      real32_t *s0=ad.s0+id;
      pos[0]=p0[OX]+dtOut*speed[0];
      pos[1]=p0[OY]+dtOut*speed[1];
      pos[2]=p0[OZ]+dtOut*speed[2];
      speed[0]=s0[OX]+dtOut*accel[0];
      speed[1]=s0[OY]+dtOut*accel[1];
      speed[2]=s0[OZ]+dtOut*accel[2];
      break;
      }
    case 3:
      {
      pos[0]=p3[OX]+dtAcc*speed[0];
      pos[1]=p3[OY]+dtAcc*speed[1];
      pos[2]=p3[OZ]+dtAcc*speed[2];
      speed[0]=s3[OX]+dtAcc*accel[0];
      speed[1]=s3[OY]+dtAcc*accel[1];
      speed[2]=s3[OZ]+dtAcc*accel[2];
      break;
      }
    }
#else
  speed[0]+=DT*accel[0];
  speed[1]+=DT*accel[1];
  speed[2]+=DT*accel[2];
# if defined USE_VERLET
    real32_t prevX=pOut[OX];
    real32_t prevY=pOut[OY];
    real32_t prevZ=pOut[OZ];
    real32_t dt2=DT*DT;
    pos[0]+=pos[0]-prevX+dt2*accel[0];
    pos[1]+=pos[1]-prevY+dt2*accel[1];
    pos[2]+=pos[2]-prevZ+dt2*accel[2];
# else
    pos[0]+=DT*speed[0];
    pos[1]+=DT*speed[1];
    pos[2]+=DT*speed[2];
# endif
#endif
pOut[OX]=pos[0];   pOut[OY]=pos[1];   pOut[OZ]=pos[2];
sOut[OX]=speed[0]; sOut[OY]=speed[1]; sOut[OZ]=speed[2];
if(spheres
#if defined USE_RK_4
   &&(ad.rk==3)
#endif
  )
  {
  prepareRendering(spheres,colors,lines,pos,state,x,y,z,id);
  }
}

void
computeBlock(int cpuCtx,
             void *data)
{
(void)data;
int idBegin=ad.threadLimits[cpuCtx];
int idEnd=ad.threadLimits[cpuCtx+1];
real32_t *spheres,*colors,*lines;
if(ad.mustRedraw)
  {
  spheres=ad.spheres;
  colors=ad.colors;
  lines=ad.lines;
  }
else
  {
  spheres=(real32_t *)0;
  colors=(real32_t *)0;
  lines=(real32_t *)0;
  }
for(int id=idBegin;id<idEnd;++id)
  {
  int z=id/(NX*NY);
  int xy=id-z*NX*NY;
  int y=xy/NX;
  int x=xy-y*NX;
  computePoint(x,y,z,spheres,colors,lines);
  }
}

void
AppliData_reset(void)
{
ad.forceX=0.0;
ad.forceY=0.0;
ad.forceZ=0.0;
ad.anchorX=NX/2;
ad.anchorY=2;
ad.anchorZ=2;
for(int z=0;z<NZ;++z)
  {
  for(int y=0;y<NY;++y)
    {
    for(int x=0;x<NX;++x)
      {
      int id=MAKE_ID(x,y,z);
      ad.s0[id+OX]=0.0;
      ad.s0[id+OY]=0.0;
      ad.s0[id+OZ]=0.0;
      ad.p1[id+OX]=ad.p0[id+OX]=(0.5f+x)/ad.delta[0];
      ad.p1[id+OY]=ad.p0[id+OY]=(0.5f+y)/ad.delta[0];
      ad.p1[id+OZ]=ad.p0[id+OZ]=(0.5f+z)/ad.delta[0];
#if defined USE_RK4
      ad.p3[id+OX]=ad.p2[id+OX]=ad.p0[id+OX];
      ad.p3[id+OY]=ad.p2[id+OY]=ad.p0[id+OY];
      ad.p3[id+OZ]=ad.p2[id+OZ]=ad.p0[id+OZ];
      ad.s3[id+OX]=ad.s2[id+OX]=ad.s1[id+OX]=ad.s0[id+OX];
      ad.s3[id+OY]=ad.s2[id+OY]=ad.s1[id+OY]=ad.s0[id+OY];
      ad.s3[id+OZ]=ad.s2[id+OZ]=ad.s1[id+OZ]=ad.s0[id+OZ];
#endif
      }
    }
  }
if(ad.gpu.prg)
  {
  CudaBuffer_sendAll(ad.gpu.p0,ad.p0);
  CudaBuffer_sendAll(ad.gpu.s0,ad.s0);
  CudaBuffer_sendAll(ad.gpu.p1,ad.p1);
#if defined USE_RK4
  CudaBuffer_sendAll(ad.gpu.s1,ad.s1);
  CudaBuffer_sendAll(ad.gpu.p2,ad.p2);
  CudaBuffer_sendAll(ad.gpu.s2,ad.s2);
  CudaBuffer_sendAll(ad.gpu.p3,ad.p3);
  CudaBuffer_sendAll(ad.gpu.s3,ad.s3);
#endif
  }
}

void
AppliData_addForce(real64_t dx,
                   real64_t dy,
                   real64_t dz)
{
ad.forceX+=100.0f*(real32_t)dx;
ad.forceY+=100.0f*(real32_t)dy;
ad.forceZ+=100.0f*(real32_t)dz;
real32_t f=sqrtf(ad.forceX*ad.forceX+
                 ad.forceY*ad.forceY+
                 ad.forceZ*ad.forceZ);
if(f>MAX_FORCE)
  {
  ad.forceX*=MAX_FORCE/f;
  ad.forceY*=MAX_FORCE/f;
  ad.forceZ*=MAX_FORCE/f;
  }
}

void
AppliData_moveAnchor(real64_t dx,
                     real64_t dy,
                     real64_t dz)
{
dx+=ad.anchorX;
dy+=ad.anchorY;
dz+=ad.anchorZ;
ad.anchorX=(int)floor(dx);
ad.anchorY=(int)floor(dy);
ad.anchorZ=(int)floor(dz);
if(dx-ad.anchorX>=0.5f) ++ad.anchorX;
if(dy-ad.anchorY>=0.5f) ++ad.anchorY;
if(dz-ad.anchorZ>=0.5f) ++ad.anchorZ;
if(ad.anchorX<1)         ad.anchorX=1;
else if(ad.anchorX>NX-2) ad.anchorX=NX-2;
if(ad.anchorY<1)         ad.anchorY=1;
else if(ad.anchorY>NY-2) ad.anchorY=NY-2;
if(ad.anchorZ<1)         ad.anchorZ=1;
else if(ad.anchorZ>NZ-2) ad.anchorZ=NZ-2;
}

void
AppliData_init(int config)
{
ad.mustQuit=false;
ad.mustRedraw=false;
ad.delta[0]=1.0f/L0;
ad.delta[1]=1.0f/(sqrtf(2.0f)*L0);
ad.delta[2]=1.0f/(sqrtf(3.0f)*L0);
ad.gpu.prg=(CudaProgram *)0;
ad.gpu.gridDimX=0;
ad.gpu.gridDimY=0;
ad.gpu.gridDimZ=0;
ad.gpu.shMem=0;
ad.threadLimits=(int *)0;
int pointCount=NX*NY*NZ;
int sz;
if(config>=0)
  {
  Cuda_init();
  const CudaDevice *dev=(const CudaDevice *)0;
  int deviceCount;
  const CudaDevice *devices=Cuda_devices(&deviceCount);
  if(config<deviceCount)
    {
    dev=devices+config;
    }
  if(!dev) DROP_DEAD("no GPU device");
#if BLOCK_DIM_Y && BLOCK_DIM_Z
  ad.gpu.gridDimX=(NX+BLOCK_DIM_X-1)/BLOCK_DIM_X;
  ad.gpu.gridDimY=(NY+BLOCK_DIM_Y-1)/BLOCK_DIM_Y;
  ad.gpu.gridDimZ=(NZ+BLOCK_DIM_Z-1)/BLOCK_DIM_Z;
#else
  ad.gpu.gridDimX=(NX*NY*NZ+BLOCK_DIM_X-1)/BLOCK_DIM_X;
#endif
#if defined SHARED_COPY
  ad.gpu.shMem=3*sizeof(real32_t)*
               (BLOCK_DIM_X+2)*(BLOCK_DIM_Y+2)*(BLOCK_DIM_Z+2);
#endif
  intptr_t sz=Path_fileSize(GPU_SOURCE_FILE);
  if(sz<0) DROP_DEAD("no file '"GPU_SOURCE_FILE"'");
  char *program=Memory_alloc(sz+0x1000,char);
  int offset=sprintf(program,
    "typedef float real32_t;\n"
    "#define NX %d\n"
    "#define NY %d\n"
    "#define NZ %d\n"
    "#define OX %d\n"
    "#define OY %d\n"
    "#define OZ %d\n"
    "#define DELTA_0 %ef\n"
    "#define DELTA_1 %ef\n"
    "#define DELTA_2 %ef\n"
    "#define DT %ef\n"
    "#define DAMP %ef\n"
    "#define MASS %ef\n"
    "#define K_ELAST %ef\n"
    "#define BLOCK_DIM_X %d\n"
    "#define BLOCK_DIM_Y %d\n"
    "#define BLOCK_DIM_Z %d\n"
    "#define GRID_DIM_X %d\n"
    "#define GRID_DIM_Y %d\n"
    "#define GRID_DIM_Z %d\n"
#if defined USE_RK4
    "#define USE_RK4\n"
#endif
#if defined USE_VERLET
    "#define USE_VERLET\n"
#endif
#if defined UNIDIMENSIONAL
    "#define UNIDIMENSIONAL\n"
#endif
#if defined GLOBAL_SPLIT_XYZ
    "#define GLOBAL_SPLIT_XYZ\n"
    "#define MAKE_ID(x,y,z) ((x)+NX*((y)+NY*(z)))\n"
#else
    "#define MAKE_ID(x,y,z) (3*((x)+NX*((y)+NY*(z))))\n"
#endif
#if defined SHARED_COPY
    "#define SHARED_COPY\n"
    "__constant__ const int shMemSize[3]={ %d, %d, %d };\n"
# if defined SHARED_SPLIT_XYZ
    "#define SHARED_SPLIT_XYZ\n"
    "#define MAKE_SHID(shx,shy,shz) \\\n"
    "        ((shx)+shMemSize[0]*((shy)+shMemSize[1]*(shz)))\n"
    "#define SHOX 0\n"
    "#define SHOY (shMemSize[0]*shMemSize[1]*shMemSize[2])\n"
    "#define SHOZ (SHOY+SHOY)\n"
# else
    "#define MAKE_SHID(shx,shy,shz) \\\n"
    "        (3*((shx)+shMemSize[0]*((shy)+shMemSize[1]*(shz))))\n"
    "#define SHOX 0\n"
    "#define SHOY 1\n"
    "#define SHOZ 2\n"
# endif
#endif
    "#line 1 \"%s\"\n",
    NX,NY,NZ,
    OX,OY,OZ,
    ad.delta[0],ad.delta[1],ad.delta[2],
    DT,DAMP,MASS,K_ELAST,
    BLOCK_DIM_X, BLOCK_DIM_Y, BLOCK_DIM_Z,
    ad.gpu.gridDimX,
#if BLOCK_DIM_Y && BLOCK_DIM_Z
    ad.gpu.gridDimY,
    ad.gpu.gridDimZ
#else
    0,0
#endif
#if defined SHARED_COPY
    ,BLOCK_DIM_X+2, BLOCK_DIM_Y+2, BLOCK_DIM_Z+2
#endif
    ,GPU_SOURCE_FILE);
  if(Path_readFile(GPU_SOURCE_FILE,program+offset,sz,0)!=sz)
    DROP_DEAD("cannot read file '"GPU_SOURCE_FILE"'");
  program[offset+sz]='\0';
#if 0
    Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    Platform_msg("%s\n",program);
    Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
#endif
  const char *options=
    "--use_fast_math "
    "";
  ad.gpu.prg=CudaProgram_new(dev,"mecaSimu",options,program,
#if defined SHARED_COPY
                             false);
#else
                             true);
#endif
  Memory_free(program);
  if(CudaProgram_buildLog(ad.gpu.prg))
    {
    Platform_msg("build log: %s\n",CudaProgram_buildLog(ad.gpu.prg));
    }
  if(CudaProgram_buildFailure(ad.gpu.prg))
    {
    DROP_DEAD("gpu code compilation failure");
    }
#if defined AVOID_RENDERING
  ad.gpu.spheres=(CudaBuffer *)0;
  ad.gpu.colors=(CudaBuffer *)0;
  ad.gpu.lines=(CudaBuffer *)0;
#else
  if(dev==Cuda_glDevice())
    {
    ad.gpu.spheres=CudaBuffer_newFromGl(dev,Rendering_sphereBuffer());
    ad.gpu.colors=CudaBuffer_newFromGl(dev,Rendering_colorBuffer());
    ad.gpu.lines=CudaBuffer_newFromGl(dev,Rendering_lineBuffer());
    ad.spheres=(real32_t *)0;
    ad.colors=(real32_t *)0;
    ad.colors=(real32_t *)0;
    }
  else
    {
    sz=pointCount*3*sizeof(real32_t);
    ad.gpu.spheres=CudaBuffer_new(dev,sz);
    ad.gpu.sphereMem=CudaLockedMem_new(false,sz);
    ad.spheres=(real32_t *)CudaLockedMem_data(ad.gpu.sphereMem);
    ad.gpu.colors=CudaBuffer_new(dev,sz);
    ad.gpu.colorMem=CudaLockedMem_new(false,sz);
    ad.colors=(real32_t *)CudaLockedMem_data(ad.gpu.colorMem);
    sz=(27*2)*3*sizeof(real32_t);
    ad.gpu.lines=CudaBuffer_new(dev,sz);
    ad.gpu.lineMem=CudaLockedMem_new(false,sz);
    ad.lines=(real32_t *)CudaLockedMem_data(ad.gpu.lineMem);
    }
#endif
  CudaLockedMem *mem;
  sz=3*pointCount*sizeof(real32_t);
  ad.gpu.p0=CudaBuffer_new(dev,sz);
  ad.gpu.p0mem=mem=CudaLockedMem_new(false,sz);
  ad.p0=(real32_t *)CudaLockedMem_data(mem);
  ad.gpu.s0=CudaBuffer_new(dev,sz);
  ad.gpu.s0mem=mem=CudaLockedMem_new(false,sz);
  ad.s0=(real32_t *)CudaLockedMem_data(mem);
  ad.gpu.p1=CudaBuffer_new(dev,sz);
  ad.gpu.p1mem=mem=CudaLockedMem_new(false,sz);
  ad.p1=(real32_t *)CudaLockedMem_data(mem);
#if defined USE_RK4
  ad.gpu.s1=CudaBuffer_new(dev,sz);
  ad.gpu.s1mem=mem=CudaLockedMem_new(false,sz);
  ad.s1=(real32_t *)CudaLockedMem_data(mem);
  ad.gpu.p2=CudaBuffer_new(dev,sz);
  ad.gpu.p2mem=mem=CudaLockedMem_new(false,sz);
  ad.p2=(real32_t *)CudaLockedMem_data(mem);
  ad.gpu.s2=CudaBuffer_new(dev,sz);
  ad.gpu.s2mem=mem=CudaLockedMem_new(false,sz);
  ad.s2=(real32_t *)CudaLockedMem_data(mem);
  ad.gpu.p3=CudaBuffer_new(dev,sz);
  ad.gpu.p3mem=mem=CudaLockedMem_new(false,sz);
  ad.p3=(real32_t *)CudaLockedMem_data(mem);
  ad.gpu.s3=CudaBuffer_new(dev,sz);
  ad.gpu.s3mem=mem=CudaLockedMem_new(false,sz);
  ad.s3=(real32_t *)CudaLockedMem_data(mem);
  ad.rk=0;
#endif
  }
else
  {
  sz=pointCount*3;
  ad.gpu.spheres=(CudaBuffer *)0;
  ad.spheres=Memory_alloc(sz,real32_t);
  ad.gpu.colors=(CudaBuffer *)0;
  ad.colors=Memory_alloc(sz,real32_t);
  sz=(27*2)*3;
  ad.gpu.lines=(CudaBuffer *)0;
  ad.lines=Memory_alloc(sz,real32_t);
  sz=pointCount*3;
  ad.gpu.p0=(CudaBuffer *)0;
  ad.p0=Memory_alloc(sz,real32_t);
  ad.gpu.s0=(CudaBuffer *)0;
  ad.s0=Memory_alloc(sz,real32_t);
  ad.gpu.p1=(CudaBuffer *)0;
  ad.p1=Memory_alloc(sz,real32_t);
#if defined USE_RK4
  ad.gpu.s1=(CudaBuffer *)0;
  ad.s1=Memory_alloc(sz,real32_t);
  ad.gpu.p2=(CudaBuffer *)0;
  ad.p2=Memory_alloc(sz,real32_t);
  ad.gpu.s2=(CudaBuffer *)0;
  ad.s2=Memory_alloc(sz,real32_t);
  ad.gpu.p3=(CudaBuffer *)0;
  ad.p3=Memory_alloc(sz,real32_t);
  ad.gpu.s3=(CudaBuffer *)0;
  ad.s3=Memory_alloc(sz,real32_t);
  ad.rk=0;
#endif
  int threadCount=Platform_cpuCtxCount();
  ad.threadLimits=Memory_alloc(threadCount+1,int);
  for(int i=0;i<=threadCount;++i)
    {
    ad.threadLimits[i]=(int)Platform_splitBegin(0,pointCount,threadCount,i);
    }
  }
AppliData_reset();
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
int config=0;
if(Platform_argc()>1)
  {
  sscanf(Platform_argv(1),"%d",&config);
  }
#if !defined AVOID_RENDERING
  Rendering_init();
#endif
AppliData_init(config);
#if defined USE_RK4
  const char *integration="Runge-Kutta-4";
#elif defined USE_VERLET
  const char *integration="Verlet";
#else
  const char *integration="Semi-implicit-Euler";
#endif
Platform_msg("Using %s integration (dt=%g s) on %d elements (%d x %d x %d)\n",
             integration,DT,NX*NY*NZ,NX,NY,NZ);
if(ad.gpu.prg)
  {
  const CudaDevice *dev=CudaProgram_device(ad.gpu.prg);
  Platform_msg("Using GPU: %s\n",
               CudaDevice_name(dev));
  Platform_msg("           multiprocessor count: %d\n",
               CudaDevice_multiprocessorCount(dev));
  Platform_msg("           warp size: %d\n",
               CudaDevice_warpSize(dev));
  Platform_msg("           compute capability: %d.%d\n",
               CudaDevice_computeCapabilityMajor(dev),
               CudaDevice_computeCapabilityMinor(dev));
  Platform_msg("           max threads per block: %d\n",
               CudaDevice_maxThreadsPerBlock(dev));
  Platform_msg("           max block dim: %d %d %d\n",
               CudaDevice_maxBlockDimX(dev),
               CudaDevice_maxBlockDimY(dev),
               CudaDevice_maxBlockDimZ(dev));
  Platform_msg("           max grid dim: %d %d %d\n",
               CudaDevice_maxGridDimX(dev),
               CudaDevice_maxGridDimY(dev),
               CudaDevice_maxGridDimZ(dev));
  Platform_msg("           max shared memory per block: %d\n",
               CudaDevice_maxSharedMemoryPerBlock(dev));
  Platform_msg("           GL sharing: %d\n",
               dev==Cuda_glDevice());
  Platform_msg("           program:\n");
  Platform_msg("             max threads per block: %d\n",
               CudaProgram_maxThreadsPerBlock(ad.gpu.prg));
  Platform_msg("             shared size bytes: %d\n",
               CudaProgram_sharedSizeBytes(ad.gpu.prg));
  Platform_msg("             const size bytes: %d\n",
               CudaProgram_constSizeBytes(ad.gpu.prg));
  Platform_msg("             local size bytes: %d\n",
               CudaProgram_localSizeBytes(ad.gpu.prg));
  Platform_msg("             num regs: %d\n",
               CudaProgram_numRegs(ad.gpu.prg));
  Platform_msg("             ptx version: %d\n",
               CudaProgram_ptxVersion(ad.gpu.prg));
  Platform_msg("             binary version: %d\n",
               CudaProgram_binaryVersion(ad.gpu.prg));
  Platform_msg("           execution:\n");
#if defined UNIDIMENSIONAL
# if BLOCK_DIM_Y && BLOCK_DIM_Z
  Platform_msg("             grid dim: %d (fake %d %d %d)\n",
               ad.gpu.gridDimX*ad.gpu.gridDimY*ad.gpu.gridDimZ,
               ad.gpu.gridDimX,ad.gpu.gridDimY,ad.gpu.gridDimZ);
  Platform_msg("             block dim: %d (fake %d %d %d)\n",
               BLOCK_DIM_X*BLOCK_DIM_Y*BLOCK_DIM_Z,
               BLOCK_DIM_X,BLOCK_DIM_Y,BLOCK_DIM_Z);
# else
  Platform_msg("             grid dim: %d\n",
               ad.gpu.gridDimX);
  Platform_msg("             block dim: %d\n",
               BLOCK_DIM_X);
# endif
#else
  Platform_msg("             grid dim: %d %d %d (%d)\n",
               ad.gpu.gridDimX,ad.gpu.gridDimY,ad.gpu.gridDimZ,
               ad.gpu.gridDimX*ad.gpu.gridDimY*ad.gpu.gridDimZ);
  Platform_msg("             block dim: %d %d %d (%d)\n",
               BLOCK_DIM_X,BLOCK_DIM_Y,BLOCK_DIM_Z,
               BLOCK_DIM_X*BLOCK_DIM_Y*BLOCK_DIM_Z);
#endif
  if((BLOCK_DIM_X>CudaDevice_maxBlockDimX(dev))||
#if BLOCK_DIM_Y && BLOCK_DIM_Z
     (BLOCK_DIM_Y>CudaDevice_maxBlockDimY(dev))||
     (BLOCK_DIM_Z>CudaDevice_maxBlockDimZ(dev))||
     (BLOCK_DIM_X*BLOCK_DIM_Y*BLOCK_DIM_Z
#else
     (BLOCK_DIM_X
#endif
      >CudaProgram_maxThreadsPerBlock(ad.gpu.prg)))
    {
    DROP_DEAD("block dim is too large");
    }
  if((ad.gpu.gridDimX>CudaDevice_maxGridDimX(dev))
#if !defined UNIDIMENSIONAL && BLOCK_DIM_Y && BLOCK_DIM_Z
     ||(ad.gpu.gridDimY>CudaDevice_maxGridDimY(dev))
     ||(ad.gpu.gridDimZ>CudaDevice_maxGridDimZ(dev)))
#else
     )
#endif
    {
    DROP_DEAD("grid dim is too large");
    }
#if defined SHARED_COPY
  Platform_msg("SHARED_COPY (%d)\n",ad.gpu.shMem);
  if((CudaProgram_sharedSizeBytes(ad.gpu.prg)+ad.gpu.shMem)>
     CudaDevice_maxSharedMemoryPerBlock(dev))
    {
    DROP_DEAD("insufficient shared memory");
    }
# if defined SHARED_SPLIT_XYZ
    printf("SHARED_SPLIT_XYZ\n");
# endif
#endif
  }
else
  {
  Platform_msg("Using %d threads\n",Platform_cpuCtxCount());
  }
# if defined GLOBAL_SPLIT_XYZ
    Platform_msg("GLOBAL_SPLIT_XYZ\n");
# endif
int64_t now=Platform_time();
int64_t lastPerf=now;
int perfCount=0;
#if !defined AVOID_RENDERING
  int64_t lastDraw=now;
#endif
while(!ad.mustQuit)
  {
  ++perfCount;
  now=Platform_time();
  int64_t dt=now-lastPerf;
  if(dt>=2000000)
    {
    real64_t rate=perfCount/(1e-6*dt);
    Platform_msg("%g Hz (%g x real time), force=%g N\n",
                 rate,rate*DT,
                 sqrt(ad.forceX*ad.forceX+
                      ad.forceY*ad.forceY+
                      ad.forceZ*ad.forceZ));
    lastPerf+=dt;
    perfCount=0;
    }
#if !defined AVOID_RENDERING
  dt=now-lastDraw;
  if(dt>40000)
    {
    Rendering_update();
    if(Rendering_visible()) ad.mustRedraw=true;
    }
#endif
  if(ad.gpu.prg)
    {
    CudaBuffer * tmpBuf=ad.gpu.p0;
    ad.gpu.p0=ad.gpu.p1;
    ad.gpu.p1=tmpBuf;
# if defined USE_RK4
    tmpBuf=ad.gpu.s0;
    ad.gpu.s0=ad.gpu.s1;
    ad.gpu.s1=tmpBuf;
# endif
    }
  real32_t *tmp=ad.p0;
  ad.p0=ad.p1;
  ad.p1=tmp;
#if defined USE_RK4
  tmp=ad.s0;
  ad.s0=ad.s1;
  ad.s1=tmp;
#endif
  if(ad.gpu.prg)
    {
    int argCount=0;
    const void *args[32];
    void *nullArg=(void *)0;
    args[argCount++]=&ad.anchorX;
    args[argCount++]=&ad.anchorY;
    args[argCount++]=&ad.anchorZ;
    args[argCount++]=&ad.forceX;
    args[argCount++]=&ad.forceY;
    args[argCount++]=&ad.forceZ;
    if(ad.mustRedraw)
      {
      if(!ad.spheres)
        {
        CudaBuffer_acquireGl(ad.gpu.spheres);
        CudaBuffer_acquireGl(ad.gpu.colors);
        CudaBuffer_acquireGl(ad.gpu.lines);
        }
      args[argCount++]=CudaBuffer_programArg(ad.gpu.spheres);
      args[argCount++]=CudaBuffer_programArg(ad.gpu.colors);
      args[argCount++]=CudaBuffer_programArg(ad.gpu.lines);
      }
    else
      {
      args[argCount++]=&nullArg;
      args[argCount++]=&nullArg;
      args[argCount++]=&nullArg;
      }
    args[argCount++]=CudaBuffer_programArg(ad.gpu.p0);
    args[argCount++]=CudaBuffer_programArg(ad.gpu.s0);
    args[argCount++]=CudaBuffer_programArg(ad.gpu.p1);
#if defined USE_RK4
    args[argCount++]=CudaBuffer_programArg(ad.gpu.s1);
    args[argCount++]=CudaBuffer_programArg(ad.gpu.p2);
    args[argCount++]=CudaBuffer_programArg(ad.gpu.s2);
    args[argCount++]=CudaBuffer_programArg(ad.gpu.p3);
    args[argCount++]=CudaBuffer_programArg(ad.gpu.s3);
    for(ad.rk=0;ad.rk<4;++ad.rk)
      {
      args[argCount]=(const void *)&ad.rk;
#else
      {
#endif
      int blockDimX,blockDimY,blockDimZ;
      int gridDimX,gridDimY,gridDimZ;
#if defined UNIDIMENSIONAL
# if BLOCK_DIM_Y && BLOCK_DIM_Z
      blockDimX=BLOCK_DIM_X*BLOCK_DIM_Y*BLOCK_DIM_Z;
      gridDimX=ad.gpu.gridDimX*ad.gpu.gridDimY*ad.gpu.gridDimZ;
# else
      blockDimX=BLOCK_DIM_X;
      gridDimX=ad.gpu.gridDimX;
# endif
      blockDimY=1;
      gridDimY=1;
      blockDimZ=1;
      gridDimZ=1;
#else
      blockDimX=BLOCK_DIM_X;
      blockDimY=BLOCK_DIM_Y;
      blockDimZ=BLOCK_DIM_Z;
      gridDimX=ad.gpu.gridDimX;
      gridDimY=ad.gpu.gridDimY;
      gridDimZ=ad.gpu.gridDimZ;
#endif
      CudaProgram_launch(ad.gpu.prg,
                         gridDimX,gridDimY,gridDimZ,
                         blockDimX,blockDimY,blockDimZ,
                         ad.gpu.shMem,args);
      }
    if(ad.mustRedraw)
      {
      if(ad.spheres)
        {
        CudaBuffer_receiveAll(ad.gpu.spheres,ad.spheres);
        CudaBuffer_receiveAll(ad.gpu.colors,ad.colors);
        CudaBuffer_receiveAll(ad.gpu.lines,ad.lines);
        }
      else
        {
        CudaBuffer_releaseGl(ad.gpu.spheres);
        CudaBuffer_releaseGl(ad.gpu.colors);
        CudaBuffer_releaseGl(ad.gpu.lines);
        }
      }
    CudaDevice_sync(CudaProgram_device(ad.gpu.prg));
    }
  else
    {
#if defined USE_RK4
    for(ad.rk=0;ad.rk<4;++ad.rk)
#endif
    Platform_parallel(&computeBlock,(void *)0);
    }
#if !defined AVOID_RENDERING
  if(ad.mustRedraw)
    {
    if(ad.spheres)
      {
      Rendering_fillBuffers(ad.spheres,ad.colors,ad.lines);
      }
    lastDraw+=dt;
    ad.mustRedraw=false;
    }
#endif
  }
if(ad.gpu.prg)
  {
  CudaProgram_delete(ad.gpu.prg);
#if !defined AVOID_RENDERING
  CudaBuffer_delete(ad.gpu.spheres);
  CudaBuffer_delete(ad.gpu.colors);
  CudaBuffer_delete(ad.gpu.lines);
  if(ad.spheres)
    {
    CudaLockedMem_delete(ad.gpu.sphereMem);
    CudaLockedMem_delete(ad.gpu.colorMem);
    CudaLockedMem_delete(ad.gpu.lineMem);
    }
#endif
  CudaBuffer_delete(ad.gpu.p0);
  CudaLockedMem_delete(ad.gpu.p0mem);
  CudaBuffer_delete(ad.gpu.s0);
  CudaLockedMem_delete(ad.gpu.s0mem);
  CudaBuffer_delete(ad.gpu.p1);
  CudaLockedMem_delete(ad.gpu.p1mem);
#if defined USE_RK4
  CudaBuffer_delete(ad.gpu.s1);
  CudaLockedMem_delete(ad.gpu.s1mem);
  CudaBuffer_delete(ad.gpu.p2);
  CudaLockedMem_delete(ad.gpu.p2mem);
  CudaBuffer_delete(ad.gpu.s2);
  CudaLockedMem_delete(ad.gpu.s2mem);
  CudaBuffer_delete(ad.gpu.p3);
  CudaLockedMem_delete(ad.gpu.p3mem);
  CudaBuffer_delete(ad.gpu.s3);
  CudaLockedMem_delete(ad.gpu.s3mem);
#endif
  }
else
  {
#if !defined AVOID_RENDERING
  Memory_free(ad.spheres);
  Memory_free(ad.colors);
  Memory_free(ad.lines);
#endif
  Memory_free(ad.p0);
  Memory_free(ad.s0);
  Memory_free(ad.p1);
#if defined USE_RK4
  Memory_free(ad.s1);
  Memory_free(ad.p2);
  Memory_free(ad.s2);
  Memory_free(ad.p3);
  Memory_free(ad.s3);
#endif
  Memory_free(ad.threadLimits);
  }
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
