//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "perf.h"

void
copy32(int count,
       const int * RESTRICT idx,
       const real32_t *in,
       real32_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r=idx[i];
  out[r]=in[r];
  }
}

void
copy64(int count,
       const int * RESTRICT idx,
       const real64_t *in,
       real64_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r=idx[i];
  out[r]=in[r];
  }
}

void
convIntTo32(int count,
            const int * RESTRICT idx,
            const int *in,
            real32_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r=idx[i];
  out[r]=(real32_t)in[r];
  }
}

void
conv32ToInt(int count,
            const int * RESTRICT idx,
            const real32_t *in,
            int *out)
{
for(int i=0;i<count;++i)
  {
  const int r=idx[i];
  out[r]=(int)in[r];
  }
}

void
convIntTo64(int count,
            const int * RESTRICT idx,
            const int *in,
            real64_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r=idx[i];
  out[r]=(real64_t)in[r];
  }
}

void
conv64ToInt(int count,
            const int * RESTRICT idx,
            const real64_t *in,
            int *out)
{
for(int i=0;i<count;++i)
  {
  const int r=idx[i];
  out[r]=(int)in[r];
  }
}

void
conv32To64(int count,
           const int * RESTRICT idx,
           const real32_t *in,
           real64_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=in[r1]+in[r2];
  }
}

void
conv64To32(int count,
           const int * RESTRICT idx,
           const real64_t *in,
           real32_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=(real32_t)(in[r1]+in[r2]);
  }
}

void
min32(int count,
      const int * RESTRICT idx,
      const real32_t *in,
      real32_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=fminf(in[r1],in[r2]);
  }
}

void
min64(int count,
      const int * RESTRICT idx,
      const real64_t *in,
      real64_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=fmin(in[r1],in[r2]);
  }
}

void
max32(int count,
      const int * RESTRICT idx,
      const real32_t *in,
      real32_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=fmaxf(in[r1],in[r2]);
  }
}

void
max64(int count,
      const int * RESTRICT idx,
      const real64_t *in,
      real64_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=fmax(in[r1],in[r2]);
  }
}

void
add32(int count,
      const int * RESTRICT idx,
      const real32_t *in,
      real32_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=in[r1]+in[r2];
  }
}

void
add64(int count,
      const int * RESTRICT idx,
      const real64_t *in,
      real64_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=in[r1]+in[r2];
  }
}

void
mul32(int count,
      const int * RESTRICT idx,
      const real32_t *in,
      real32_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=in[r1]*in[r2];
  }
}

void
mul64(int count,
      const int * RESTRICT idx,
      const real64_t *in,
      real64_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=in[r1]*in[r2];
  }
}

void
div32(int count,
      const int * RESTRICT idx,
      const real32_t *in,
      real32_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=in[r1]/in[r2];
  }
}

void
div64(int count,
      const int * RESTRICT idx,
      const real64_t *in,
      real64_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r1=idx[i];
  const int r2=idx[count-1-i];
  out[r1]=in[r1]/in[r2];
  }
}

void
classic32(int count,
          const int * RESTRICT idx,
          const real32_t *in,
          real32_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r=idx[i];
  out[r]=1.0f/sqrtf(in[r]);
  }
}

void
approx32(int count,
         const int * RESTRICT idx,
         const real32_t *in,
         real32_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r=idx[i];
  out[r]=approxInvSqrt32(in[r]);
  }
}

void
classic64(int count,
          const int * RESTRICT idx,
          const real64_t *in,
          real64_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r=idx[i];
  out[r]=1.0/sqrt(in[r]);
  }
}

void
approx64(int count,
         const int * RESTRICT idx,
         const real64_t *in,
         real64_t *out)
{
for(int i=0;i<count;++i)
  {
  const int r=idx[i];
  out[r]=approxInvSqrt64(in[r]);
  }
}

//----------------------------------------------------------------------------
