//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/platform.h"

void
showMemCtx(int memCtx,
           int indent,
           bool showRedundancy)
{
for(int i=0;i<indent;++i)
  {
  Platform_msg(" ");
  }
indent+=1+Platform_msg("Node[%d]",memCtx);
int parent=Platform_memParent(memCtx);
int numaCtx=Platform_memNumaCtx(memCtx);
if(Platform_numaId(numaCtx)!=-1)
  {
  if(showRedundancy||(parent==-1)||(Platform_memNumaCtx(parent)!=numaCtx))
    {
    Platform_msg(" Numa[%d](%d)",
                 numaCtx,Platform_numaId(numaCtx));
    }
  }
for(int i=Platform_maxCacheLevel();i>0;--i)
  {
  if(Platform_memCacheSize(memCtx,i))
    {
    if(showRedundancy||(parent==-1)||!Platform_memCacheSize(parent,i))
      {
      Platform_msg(" L%d(%d/%d)",
                   i,
                   Platform_memCacheSize(memCtx,i),
                   Platform_memCacheLineSize(memCtx,i));
      }
    }
  }
if(showRedundancy||!Platform_memChildCount(memCtx))
  {
  Platform_msg(" {");
  for(int i=0,i_end=Platform_memCpuCount(memCtx);i<i_end;++i)
    {
    int cpuCtx=Platform_memFirstCpu(memCtx)+i;
    Platform_msg(" %d",Platform_cpuId(cpuCtx));
    }
  Platform_msg(" }");
  }
Platform_msg("\n");
for(int i=0,i_end=Platform_memChildCount(memCtx);i<i_end;++i)
  {
  showMemCtx(Platform_memFirstChild(memCtx)+i,indent,showRedundancy);
  }
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
bool showRedundancy=false;
bool showRoundTrip=false;
for(int i=1,i_end=Platform_argc();i<i_end;++i)
  {
  if(!strcmp(Platform_argv(i),"-r"))
    {
    showRedundancy=true;
    }
  else if(!strcmp(Platform_argv(i),"-t"))
    {
    showRoundTrip=true;
    }
  }
Platform_msg("memory size: %"F64"d\n",Platform_memorySize());
Platform_msg("standard page size: %"FPTR"d\n",Platform_standardPageSize());
Platform_msg("huge page size: %"FPTR"d\n",Platform_hugePageSize());
Platform_msg("max cache level: %d\n",Platform_maxCacheLevel());
Platform_msg("max cache line size: %d\n",Platform_maxCacheLineSize());
Platform_msg("number of numa contexts: %d\n",Platform_numaCtxCount());
Platform_msg("max numa id: %d\n",Platform_maxNumaId());
Platform_msg("number of cpu contexts: %d\n",Platform_cpuCtxCount());
Platform_msg("max cpu id: %d\n",Platform_maxCpuId());
Platform_msg("number of memory contexts: %d\n",Platform_memCtxCount());
showMemCtx(Platform_rootMemCtx(),0,showRedundancy);
if(showRoundTrip)
  {
  int cpuCount=Platform_cpuCtxCount();
  for(int c=0;c<cpuCount;++c)
    {
    Platform_msg("cpu %d:",c);
    const NarrowCpuCtx *trip=Platform_cpuRoundTrip(c);
    for(int i=0;i<cpuCount;++i)
      {
      Platform_msg(" %d",(int)trip[i]);
      }
    Platform_msg("\n");
    }
  }
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
