//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "perf.h"

void
test_32_64(void)
{
const int cpuCtx=Platform_cpuCtx();
const int block=512;
real32_t *in32=Memory_alloc(block,real32_t);
real32_t *out32=Memory_alloc(block,real32_t);
real64_t *in64=Memory_alloc(block,real64_t);
real64_t *out64=Memory_alloc(block,real64_t);
int *inInt=Memory_alloc(block,int);
int *outInt=Memory_alloc(block,int);
int *idx=Memory_alloc(block,int);
for(int i=0;i<block;++i)
  {
  in32[i]=(real32_t)(in64[i]=Random_real64(cpuCtx));
  out32[i]=(real32_t)(out64[i]=0.0);
  inInt[i]=Random_int(cpuCtx,block);
  outInt[i]=0;
  idx[i]=-1;
  }
for(int i=0;i<block;++i)
  {
  int r=Random_int(cpuCtx,block);
  while(idx[r]!=-1)
    {
    r=(r+1)%block;
    }
  idx[r]=i;
  }
const int iter1=64;
const int iter2=65536;

Platform_msg("real32_t=real32_t:\n");
real64_t copy32Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    copy32(block,idx,in32,out32);
    }
  copy32Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             copy32Dt);

Platform_msg("real64_t=real64_t:\n");
real64_t copy64Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    copy64(block,idx,in64,out64);
    }
  copy64Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             copy64Dt);

Platform_msg("real32_t=int:\n");
real64_t convIntTo32Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    convIntTo32(block,idx,inInt,out32);
    }
  convIntTo32Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             convIntTo32Dt);

Platform_msg("int=real32_t:\n");
real64_t conv32ToIntDt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    conv32ToInt(block,idx,in32,outInt);
    }
  conv32ToIntDt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             conv32ToIntDt);

Platform_msg("real64_t=int:\n");
real64_t convIntTo64Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    convIntTo64(block,idx,inInt,out64);
    }
  convIntTo64Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             convIntTo64Dt);

Platform_msg("int=real64_t:\n");
real64_t conv64ToIntDt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    conv64ToInt(block,idx,in64,outInt);
    }
  conv64ToIntDt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             conv64ToIntDt);

Platform_msg("real32_t=real64_t:\n");
real64_t conv64To32Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    conv64To32(block,idx,in64,out32);
    }
  conv64To32Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             conv64To32Dt);

Platform_msg("real64_t=real32_t:\n");
real64_t conv32To64Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    conv32To64(block,idx,in32,out64);
    }
  conv32To64Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             conv32To64Dt);

Platform_msg("fminf(real32_t,real32_t):\n");
real64_t min32Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    min32(block,idx,in32,out32);
    }
  min32Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             min32Dt);

Platform_msg("fmin(real64_t,real64_t):\n");
real64_t min64Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    min64(block,idx,in64,out64);
    }
  min64Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x fminf(real32_t,real32_t)\n\n",
             min64Dt,
             min32Dt/min64Dt);

Platform_msg("fmaxf(real32_t,real32_t):\n");
real64_t max32Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    max32(block,idx,in32,out32);
    }
  max32Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             max32Dt);

Platform_msg("fmax(real64_t,real64_t):\n");
real64_t max64Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    max64(block,idx,in64,out64);
    }
  max64Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x fmaxf(real32_t,real32_t)\n\n",
             max64Dt,
             max32Dt/max64Dt);

Platform_msg("real32_t+real32_t:\n");
real64_t add32Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    add32(block,idx,in32,out32);
    }
  add32Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             add32Dt);

Platform_msg("real64_t+real64_t:\n");
real64_t add64Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    add64(block,idx,in64,out64);
    }
  add64Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x real32_t+real32_t\n\n",
             add64Dt,
             add32Dt/add64Dt);

Platform_msg("real32_t*real32_t:\n");
real64_t mul32Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    mul32(block,idx,in32,out32);
    }
  mul32Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             mul32Dt);

Platform_msg("real64_t*real64_t:\n");
real64_t mul64Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    mul64(block,idx,in64,out64);
    }
  mul64Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x real32_t*real32_t\n\n",
             mul64Dt,
             mul32Dt/mul64Dt);

Platform_msg("real32_t/real32_t:\n");
real64_t div32Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    div32(block,idx,in32,out32);
    }
  div32Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             div32Dt);

Platform_msg("real64_t/real64_t:\n");
real64_t div64Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    div64(block,idx,in64,out64);
    }
  div64Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x real32_t/real32_t\n\n",
             div64Dt,
             div32Dt/div64Dt);

Platform_msg("1.0f/sqrtf(x):\n");
real64_t classic32Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    classic32(block,idx,in32,out32);
    }
  classic32Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             classic32Dt);

Platform_msg("approxInvSqrt32(x):\n");
real64_t approx32Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    approx32(block,idx,in32,out32);
    }
  approx32Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x 1.0f/sqrtf(x)\n\n",
             approx32Dt,
             classic32Dt/approx32Dt);

Platform_msg("1.0/sqrt(x):\n");
real64_t classic64Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    classic64(block,idx,in64,out64);
    }
  classic64Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x 1.0f/sqrtf(x)\n\n",
             classic64Dt,
             classic32Dt/classic64Dt);

Platform_msg("approxInvSqrt64(x):\n");
real64_t approx64Dt=0.0;
for(int i=0;i<iter1;++i)
  {
  int64_t t0=Platform_time();
  for(int j=0;j<iter2;++j)
    {
    approx64(block,idx,in64,out64);
    }
  approx64Dt+=(Platform_time()-t0)/(iter1*(real64_t)iter2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x 1.0/sqrt(x)\n"
             "  %g x approxInvSqrt32(x)\n"
             "  %g x 1.0f/sqrtf(x)\n\n",
             approx64Dt,
             classic64Dt/approx64Dt,
             approx32Dt/approx64Dt,
             classic32Dt/approx64Dt);

Memory_free(in32);
Memory_free(in64);
Memory_free(out32);
Memory_free(out64);
Memory_free(idx);
}

#define RSTR RESTRICT
#if !defined RSTR
# define RSTR
#endif

// typedef uint64_t index_t;
// typedef uint32_t index_t;
// typedef int64_t index_t;
typedef int32_t index_t;

typedef struct _Quadratic
{
real32_t a,b,c;
real32_t d;
real32_t x0,x1;
} Quadratic;

typedef struct _Quadratics
{
real32_t *a,*b,*c;
real32_t *d;
real32_t *x0,*x1;
} Quadratics;

Quadratic *aosQuads;
Quadratics soaQuads;

void
aosSolve(Quadratic * RSTR v,
         index_t beginIndex,
         index_t endIndex)
{
TRANSPROG_VECTORIZE
for(index_t i=beginIndex;i<endIndex;++i)
  {
  real32_t a=v[i].a;
  real32_t b=v[i].b;
  real32_t c=v[i].c;
  real32_t d=b*b-4.0f*a*c;
  v[i].d=d;
  // if(d>=0.0f)
    {
    real32_t sqrtD=sqrtf(d);
    real32_t div2a=0.5f/a;
    real32_t x0=(-b-sqrtD)*div2a;
    real32_t x1=(-b+sqrtD)*div2a;
    v[i].x0=x0;
    v[i].x1=x1;
    }
  }
}

void
soaSolve(const real32_t * RSTR va,
         const real32_t * RSTR vb,
         const real32_t * RSTR vc,
         real32_t * RSTR vd,
         real32_t * RSTR vx0,
         real32_t * RSTR vx1,
         index_t beginIndex,
         index_t endIndex)
{
TRANSPROG_VECTORIZE
for(index_t i=beginIndex;i<endIndex;++i)
  {
  real32_t a=va[i];
  real32_t b=vb[i];
  real32_t c=vc[i];
  real32_t d=b*b-4.0f*a*c;
  vd[i]=d;
  // if(d>=0.0f)
    {
    real32_t sqrtD=sqrtf(d);
    real32_t div2a=0.5f/a;
    real32_t x0=(-b-sqrtD)*div2a;
    real32_t x1=(-b+sqrtD)*div2a;
    vx0[i]=x0;
    vx1[i]=x1;
    }
  }
}

void
aosTask(int cpuCtx,
        void *data)
{
ASSERT_CURRENT_CPUCTX(cpuCtx);
index_t *indices=(index_t *)data;
index_t beginIndex=indices[cpuCtx+0];
index_t endIndex=indices[cpuCtx+1];
for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
  {
  aosSolve(aosQuads,beginIndex,endIndex);
  }
}

void
soaTask(int cpuCtx,
        void *data)
{
ASSERT_CURRENT_CPUCTX(cpuCtx);
index_t *indices=(index_t *)data;
index_t beginIndex=indices[cpuCtx+0];
index_t endIndex=indices[cpuCtx+1];
for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
  {
  soaSolve(soaQuads.a,soaQuads.b,soaQuads.c,
           soaQuads.d,soaQuads.x0,soaQuads.x1,
           beginIndex,endIndex);
  }
}

void
testSimd(void)
{
const index_t QUAD_COUNT=1<<17;
const index_t ITER1=64;
const index_t ITER2=128;
aosQuads=Memory_alloc(QUAD_COUNT,Quadratic);
soaQuads.a=Memory_alloc(QUAD_COUNT,real32_t);
soaQuads.b=Memory_alloc(QUAD_COUNT,real32_t);
soaQuads.c=Memory_alloc(QUAD_COUNT,real32_t);
soaQuads.d=Memory_alloc(QUAD_COUNT,real32_t);
soaQuads.x0=Memory_alloc(QUAD_COUNT,real32_t);
soaQuads.x1=Memory_alloc(QUAD_COUNT,real32_t);

int curCtx=Platform_cpuCtx();
for(index_t i=0;i<QUAD_COUNT;++i)
  {
  real32_t a=100.0f*Random_real32(curCtx)-50.0f;
  real32_t b=100.0f*Random_real32(curCtx)-50.0f;
  real32_t c=100.0f*Random_real32(curCtx)-50.0f;
  aosQuads[i].a=soaQuads.a[i]=a;
  aosQuads[i].b=soaQuads.b[i]=b;
  aosQuads[i].c=soaQuads.c[i]=c;
  aosQuads[i].d=soaQuads.d[i]=-1.0f;
  aosQuads[i].x0=soaQuads.x0[i]=0.0f;
  aosQuads[i].x1=soaQuads.x1[i]=0.0f;
  }

index_t *indices=Memory_alloc(Platform_cpuCtxCount()+1,index_t);
for(index_t i=0,i_last=(index_t)Platform_cpuCtxCount();i<=i_last;++i)
  {
  indices[i]=(index_t)Platform_splitBegin(0,QUAD_COUNT,i_last,i);
  }

Platform_msg("serial array-of-structures:\n");
real64_t serialAosDt=0.0;
for(index_t i=0;i<ITER1;++i)
  {
  int64_t t0=Platform_time();
  for(index_t j=0;j<ITER2;++j)
    {
    aosSolve(aosQuads,0,QUAD_COUNT);
    }
  serialAosDt+=(Platform_time()-t0)/(ITER1*(real64_t)ITER2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n\n",
             serialAosDt);

Platform_msg("serial structure-of-arrays:\n");
real64_t serialSoaDt=0.0;
for(index_t i=0;i<ITER1;++i)
  {
  int64_t t0=Platform_time();
  for(index_t j=0;j<ITER2;++j)
    {
    soaSolve(soaQuads.a,soaQuads.b,soaQuads.c,
             soaQuads.d,soaQuads.x0,soaQuads.x1,
             0,QUAD_COUNT);
    }
  serialSoaDt+=(Platform_time()-t0)/(ITER1*(real64_t)ITER2);
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x array-of-structures\n\n",
             serialSoaDt,
             serialAosDt/serialSoaDt);

Platform_msg("%d-thread array-of-structures:\n",
             Platform_cpuCtxCount());
real64_t parallelAosDt=0;
for(index_t i=0;i<ITER1;++i)
  {
  int64_t t0=Platform_time();
  for(index_t j=0;j<ITER2;++j)
    {
    Platform_parallel(aosTask,indices);
    }
  parallelAosDt+=(Platform_time()-t0)/
                 (ITER1*(real64_t)ITER2*Platform_cpuCtxCount());
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x serial\n\n",
             parallelAosDt,
             serialAosDt/parallelAosDt);

Platform_msg("%d-thread structure-of-arrays:\n",
             Platform_cpuCtxCount());
real64_t parallelSoaDt=0;
for(index_t i=0;i<ITER1;++i)
  {
  int64_t t0=Platform_time();
  for(index_t j=0;j<ITER2;++j)
    {
    Platform_parallel(soaTask,indices);
    }
  parallelSoaDt+=(Platform_time()-t0)/
                 (ITER1*(real64_t)ITER2*Platform_cpuCtxCount());
  Platform_msg(".");
  }
Platform_msg(" %g us\n"
             "  %g x parallel array-of-structures\n"
             "  %g x serial structure-of-arrays\n"
             "  %g x serial array-of-structures\n\n",
             parallelSoaDt,
             parallelAosDt/parallelSoaDt,
             serialSoaDt/parallelSoaDt,
             serialAosDt/parallelSoaDt);

Memory_free(aosQuads);
Memory_free(soaQuads.a);
Memory_free(soaQuads.b);
Memory_free(soaQuads.c);
Memory_free(soaQuads.d);
Memory_free(soaQuads.x0);
Memory_free(soaQuads.x1);
Memory_free(indices);
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
test_32_64();
testSimd();
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
