//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSGL_GLVIEW_H
#define TRANSGL_GLVIEW_H 1

#include "TransProg/gfxUtils.h"

#include "TransGl/coreGl.h"

typedef struct _GlView GlView;

typedef
bool // true: keep view    false: close view
(*GlViewCallback)(GlView *view);

struct _GlView
{
char title[0x100];
GfxCursor cursor;
GfxState state;
bool keyRepeat;
GlViewCallback callback;
void *data;
void *specific;
};

TRANSGL_API
void
GlView_init(void);

TRANSGL_API
GfxStereoMode
GlView_stereoMode(void);

TRANSGL_API
int // screen width
GlView_screenWidth(void);

TRANSGL_API
int // screen height
GlView_screenHeight(void);

TRANSGL_API
int // events occured
GlView_processEvents(int64_t usecTimeout);

TRANSGL_API
void
GlView_sync(void);

TRANSGL_API
GlView * // created view
GlView_open(int initX,
            int initY,
            int initWidth,
            int initHeight,
            bool borderless,
            GlViewCallback callback,
            void *data);

TRANSGL_API
void
GlView_setTitle(GlView *view,
                const char *title);

TRANSPROG_INLINE
const char * // view title
GlView_title(const GlView *view)
{
return view->title;
}

TRANSGL_API
void
GlView_setCursor(GlView *view,
                 GfxCursor cursor);

TRANSPROG_INLINE
GfxCursor // view cursor
GlView_cursor(const GlView *view)
{
return view->cursor;
}

TRANSPROG_INLINE
void
GlView_setKeyRepeat(GlView *view,
                    bool keyRepeat)
{
view->keyRepeat=keyRepeat;
}

TRANSPROG_INLINE
bool // keyboard can repeat
GlView_keyRepeat(const GlView *view)
{
return view->keyRepeat;
}

TRANSGL_API
void
GlView_postRedraw(GlView *view);

TRANSPROG_INLINE
const GfxState * // view state
GlView_state(const GlView *view)
{
return &view->state;
}

TRANSPROG_INLINE
void * // view data
GlView_data(const GlView *view)
{
return view->data;
}

TRANSGL_API
void 
GlView_makeCurrentGlTarget(GlView *view);

TRANSGL_API
const intptr_t *
GlView_glContext(void);

#endif // TRANSGL_GLVIEW_H

//----------------------------------------------------------------------------
