//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_DYNAMICCOMPILER_H
#define TRANSPROG_DYNAMICCOMPILER_H 1

#include "TransProg/strVector.h"
#include "TransProg/text.h"
#include "TransProg/plugin.h"

typedef struct _DynamicCompiler
{
StrVector cc;
StrVector cflags_only_I;
StrVector cflags_only_D;
StrVector cflags_only_other;
StrVector libs_only_L;
StrVector libs_only_l;
StrVector libs_only_other;
StrVector oflags;
Text buildLog;
} DynamicCompiler;

TRANSPROG_API
void
DynamicCompiler_init(DynamicCompiler *dc);

TRANSPROG_INLINE
DynamicCompiler *
DynamicCompiler_new(void)
{
DynamicCompiler *dc=Memory_alloc(1,DynamicCompiler);
DynamicCompiler_init(dc);
return dc;
}

TRANSPROG_API
void
DynamicCompiler_destroy(DynamicCompiler *dc);

TRANSPROG_INLINE
void
DynamicCompiler_delete(DynamicCompiler *dc)
{
DynamicCompiler_destroy(dc);
Memory_free(dc);
}

TRANSPROG_INLINE
StrVector * // compiler command
DynamicCompiler_cc(DynamicCompiler *dc)
{
return &dc->cc;
}

TRANSPROG_INLINE
StrVector * // compiler options controling header search paths
DynamicCompiler_cflags_only_I(DynamicCompiler *dc)
{
return &dc->cflags_only_I;
}

TRANSPROG_INLINE
StrVector * // compiler options controling macro definitions
DynamicCompiler_cflags_only_D(DynamicCompiler *dc)
{
return &dc->cflags_only_D;
}

TRANSPROG_INLINE
StrVector * // other compiler options
DynamicCompiler_cflags_only_other(DynamicCompiler *dc)
{
return &dc->cflags_only_other;
}

TRANSPROG_INLINE
StrVector * // linker options controling library search paths
DynamicCompiler_libs_only_L(DynamicCompiler *dc)
{
return &dc->libs_only_L;
}

TRANSPROG_INLINE
StrVector * // linker options controling used libraries
DynamicCompiler_libs_only_l(DynamicCompiler *dc)
{
return &dc->libs_only_l;
}

TRANSPROG_INLINE
StrVector * // other linker options
DynamicCompiler_libs_only_other(DynamicCompiler *dc)
{
return &dc->libs_only_other;
}

TRANSPROG_INLINE
StrVector * // linker options controling output format and file name
DynamicCompiler_oflags(DynamicCompiler *dc)
{
return &dc->oflags;
}

TRANSPROG_API
PluginHandle // built plugin or 0
DynamicCompiler_build(DynamicCompiler *dc,
                      const char *sourceCode);

TRANSPROG_INLINE
const char * // build log or 0
DynamicCompiler_buildLog(const DynamicCompiler *dc)
{
return Text_length(&dc->buildLog) ? Text_get(&dc->buildLog)
                                  : (const char *)0;
}

#endif // TRANSPROG_DYNAMICCOMPILER_H

//----------------------------------------------------------------------------
