#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

content='//'+'-'*76+r'''
// Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
// 
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef GUARD_MACRO
#define GUARD_MACRO 1
#line 44 THIS_FILE

#include "TransProg/memory.h"

SPECIFIC_CODE

typedef ELEM_TYPE RBTREE_TYPE_ElemType;

typedef int (*RBTREE_TYPE_CmpFnct)(const RBTREE_TYPE_ElemType *,
                                   const RBTREE_TYPE_ElemType *);

#if !defined RBTREE_TYPE_NO_DEFAULT_CMP
  TRANSPROG_INLINE
  int // -1  0  1
  RBTREE_TYPE_defaultCmpUpFnct(const RBTREE_TYPE_ElemType *a,
                               const RBTREE_TYPE_ElemType *b)
  {
  return *a==*b ? 0 : *a<*b ? -1 : 1;
  }
  
  TRANSPROG_INLINE
  int // -1  0  1
  RBTREE_TYPE_defaultCmpDownFnct(const RBTREE_TYPE_ElemType *a,
                                 const RBTREE_TYPE_ElemType *b)
  {
  return *a==*b ? 0 : *a<*b ? 1 : -1;
  }
#endif

typedef void (*RBTREE_TYPE_VisitFnct)(const RBTREE_TYPE_ElemType *,void *);

typedef struct _RBTREE_TYPENode
{
RBTREE_TYPE_ElemType data;
struct _RBTREE_TYPENode *left;
struct _RBTREE_TYPENode *right;
intptr_t parent_red; // lower bit is red !
} RBTREE_TYPENode;

typedef struct _RBTREE_TYPE
{
RBTREE_TYPENode *root;
RBTREE_TYPE_CmpFnct cmp;
intptr_t size;
} RBTREE_TYPE;

TRANSPROG_INLINE
bool
_RBTREE_TYPE_red(const struct _RBTREE_TYPENode *n)
{
return n->parent_red&1;
}

TRANSPROG_INLINE
void
_RBTREE_TYPE_setRed(struct _RBTREE_TYPENode *n,
                    bool red)
{
n->parent_red=(n->parent_red&~1)|(intptr_t)red;
}

TRANSPROG_INLINE
void
_RBTREE_TYPE_toRed(struct _RBTREE_TYPENode *n)
{
n->parent_red|=1;
}

TRANSPROG_INLINE
void
_RBTREE_TYPE_toBlack(struct _RBTREE_TYPENode *n)
{
n->parent_red&=~1;
}

TRANSPROG_INLINE
struct _RBTREE_TYPENode *
_RBTREE_TYPE_parent(const struct _RBTREE_TYPENode *n)
{
return (struct _RBTREE_TYPENode *)(n->parent_red&~1);
}

TRANSPROG_INLINE
void
_RBTREE_TYPE_setParent(struct _RBTREE_TYPENode *n,
                       struct _RBTREE_TYPENode *p)
{
n->parent_red=(n->parent_red&1)|(intptr_t)p;
}

TRANSPROG_INLINE
void
RBTREE_TYPE_init(RBTREE_TYPE *t,
                 RBTREE_TYPE_CmpFnct cmp)
{
t->root=(RBTREE_TYPENode *)0;
#if !defined RBTREE_TYPE_NO_DEFAULT_CMP
  if(!cmp)
    {
    cmp=RBTREE_TYPE_defaultCmpUpFnct;
    }
#endif
t->cmp=cmp;
t->size=0;
}

TRANSPROG_INLINE
RBTREE_TYPE *
RBTREE_TYPE_new(RBTREE_TYPE_CmpFnct cmp)
{
RBTREE_TYPE *t=Memory_alloc(1,RBTREE_TYPE);
RBTREE_TYPE_init(t,cmp);
return t;
}

TRANSPROG_INLINE
void
_RBTREE_TYPE_destroyNode(RBTREE_TYPENode *n)
{
if(n)
  {
  _RBTREE_TYPE_destroyNode(n->left);
  _RBTREE_TYPE_destroyNode(n->right);
  Memory_free(n);
  }
}

TRANSPROG_INLINE
void
RBTREE_TYPE_destroy(RBTREE_TYPE *t)
{
_RBTREE_TYPE_destroyNode(t->root);
}

TRANSPROG_INLINE
void
RBTREE_TYPE_delete(RBTREE_TYPE *t)
{
RBTREE_TYPE_destroy(t);
Memory_free(t);
}

TRANSPROG_INLINE
void
_RBTREE_TYPE_nodeNumaUsage(RBTREE_TYPENode *n,
                           intptr_t *numaUsage)
{
if(n)
  {
  _RBTREE_TYPE_nodeNumaUsage(n->left,numaUsage);
  _RBTREE_TYPE_nodeNumaUsage(n->right,numaUsage);
  Memory_numaUsage(n,numaUsage);
  }
}

TRANSPROG_INLINE
void
RBTREE_TYPE_numaUsage(const RBTREE_TYPE *t,
                      intptr_t *numaUsage)
{
_RBTREE_TYPE_nodeNumaUsage(t->root,numaUsage);
}

TRANSPROG_INLINE
intptr_t
RBTREE_TYPE_size(const RBTREE_TYPE *t)
{
return t->size;
}

TRANSPROG_INLINE
RBTREE_TYPE_ElemType
RBTREE_TYPE_get(const RBTREE_TYPENode *n)
{
return n->data;
}

TRANSPROG_INLINE
void
RBTREE_TYPE_set(RBTREE_TYPENode *n,
                RBTREE_TYPE_ElemType element)
{
n->data=element;
}

TRANSPROG_INLINE
RBTREE_TYPENode *
RBTREE_TYPE_find(const RBTREE_TYPE *t,
                 RBTREE_TYPE_ElemType element)
{
RBTREE_TYPENode *n=t->root;
while(n)
  {
  int r=(*t->cmp)(&element,&n->data);
  if(r)
    {
    n=r<0 ? n->left : n->right;
    }
  else
    {
    return n;
    }
  }
return (RBTREE_TYPENode *)0;
}

TRANSPROG_INLINE
RBTREE_TYPENode *
RBTREE_TYPE_first(const RBTREE_TYPE *t)
{
RBTREE_TYPENode *n=t->root;
if(n)
  {
  RBTREE_TYPENode *l;
  while((l=n->left))
    {
    n=l;
    }
  }
return n;
}

TRANSPROG_INLINE
RBTREE_TYPENode *
RBTREE_TYPE_last(const RBTREE_TYPE *t)
{
RBTREE_TYPENode *n=t->root;
if(n)
  {
  RBTREE_TYPENode *r;
  while((r=n->right))
    {
    n=r;
    }
  }
return n;
}

TRANSPROG_INLINE
RBTREE_TYPENode *
RBTREE_TYPE_next(RBTREE_TYPENode *n)
{
if(n->right)
  {
  n=n->right;
  RBTREE_TYPENode *l;
  while((l=n->left))
    {
    n=l;
    }
  return n;
  }
else
  {
  RBTREE_TYPENode *p;
  while((p=_RBTREE_TYPE_parent(n))&&(n==p->right))
    {
    n=p;
    }
  return p;
  }
}

TRANSPROG_INLINE
RBTREE_TYPENode *
RBTREE_TYPE_previous(RBTREE_TYPENode *n)
{
if(n->left)
  {
  n=n->left;
  RBTREE_TYPENode *r;
  while((r=n->right))
    {
    n=r;
    }
  return n;
  }
else
  {
  RBTREE_TYPENode *p;
  while((p=_RBTREE_TYPE_parent(n))&&(n==p->left))
    {
    n=p;
    }
  return p;
  }
}

TRANSPROG_INLINE
void
_RBTREE_TYPE_forEachNode(const RBTREE_TYPENode *n,
                         RBTREE_TYPE_VisitFnct visitor,
                         void *arg)
{
if(n)
  {
  _RBTREE_TYPE_forEachNode(n->left,visitor,arg);
  (*visitor)(&n->data,arg);
  _RBTREE_TYPE_forEachNode(n->right,visitor,arg);
  }
}

TRANSPROG_INLINE
void
RBTREE_TYPE_forEach(const RBTREE_TYPE *t,
                    RBTREE_TYPE_VisitFnct visitor,
                    void *arg)
{
_RBTREE_TYPE_forEachNode(t->root,visitor,arg);
}

TRANSPROG_INLINE
void
_RBTREE_TYPE_rotateLeft(RBTREE_TYPE *t,
                        RBTREE_TYPENode *n)
{
RBTREE_TYPENode *r=n->right;
RBTREE_TYPENode *p=_RBTREE_TYPE_parent(n);
if((n->right=r->left))
  {
  _RBTREE_TYPE_setParent(r->left,n);
  }
r->left=n;
_RBTREE_TYPE_setParent(r,p);
if(p)
  {
  if(n==p->left)
    {
    p->left=r;
    }
  else
    {
    p->right=r;
    }
  }
else
  {
  t->root=r;
  }
_RBTREE_TYPE_setParent(n,r);
}

TRANSPROG_INLINE
void
_RBTREE_TYPE_rotateRight(RBTREE_TYPE *t,
                         RBTREE_TYPENode *n)
{
RBTREE_TYPENode *l=n->left;
RBTREE_TYPENode *p=_RBTREE_TYPE_parent(n);
if((n->left=l->right))
  {
  _RBTREE_TYPE_setParent(l->right,n);
  }
l->right=n;
_RBTREE_TYPE_setParent(l,p);
if(p)
  {
  if(n==p->right)
    {
    p->right=l;
    }
  else
    {
    p->left=l;
    }
  }
else
  {
  t->root=l;
  }
_RBTREE_TYPE_setParent(n,l);
}

TRANSPROG_INLINE
const RBTREE_TYPENode *
RBTREE_TYPE_insert(RBTREE_TYPE *t,
                   RBTREE_TYPE_ElemType element)
{
RBTREE_TYPENode *p=(RBTREE_TYPENode *)0;
RBTREE_TYPENode *n=t->root;
int r=0;
while(n)
  {
  r=(*t->cmp)(&element,&n->data);
  if(r)
    {
    p=n;
    n=r<0 ? n->left : n->right;
    }
  else
    {
    n->data=element;
    return n;
    }
  }
++(t->size);
n=Memory_alloc(1,RBTREE_TYPENode);
n->data=element;
_RBTREE_TYPE_toRed(n);
n->left=(RBTREE_TYPENode *)0;
n->right=(RBTREE_TYPENode *)0;
_RBTREE_TYPE_setParent(n,p);
RBTREE_TYPENode *result=n;
if(p)
  {
  if(r<0)
    {
    p->left=n;
    }
  else
    {
    p->right=n;
    }
  }
else
  {
  t->root=n;
  }
while((p=_RBTREE_TYPE_parent(n))&&_RBTREE_TYPE_red(p))
  {
  RBTREE_TYPENode *g=_RBTREE_TYPE_parent(p);
  if(p==g->left)
    {
    RBTREE_TYPENode *u=g->right;
    if(u&&_RBTREE_TYPE_red(u))
      {
      _RBTREE_TYPE_toBlack(u);
      _RBTREE_TYPE_toBlack(p);
      _RBTREE_TYPE_toRed(g);
      n=g;
      }
    else
      {
      if(p->right==n)
        {
        _RBTREE_TYPE_rotateLeft(t,p);
        u=p; p=n; n=u;
        }
      _RBTREE_TYPE_toBlack(p);
      _RBTREE_TYPE_toRed(g);
      _RBTREE_TYPE_rotateRight(t,g);
      }
    }
  else
    {
    RBTREE_TYPENode *u=g->left;
    if(u&&_RBTREE_TYPE_red(u))
      {
      _RBTREE_TYPE_toBlack(u);
      _RBTREE_TYPE_toBlack(p);
      _RBTREE_TYPE_toRed(g);
      n=g;
      }
    else
      {
      if(p->left==n)
        {
        _RBTREE_TYPE_rotateRight(t,p);
        u=p; p=n; n=u;
        }
      _RBTREE_TYPE_toBlack(p);
      _RBTREE_TYPE_toRed(g);
      _RBTREE_TYPE_rotateLeft(t,g);
      }
    }
  }
_RBTREE_TYPE_toBlack(t->root);
return result;
}

TRANSPROG_INLINE
void
RBTREE_TYPE_eraseNode(RBTREE_TYPE *t,
                      RBTREE_TYPENode *n)
{
RBTREE_TYPENode *destroyed=n;
RBTREE_TYPENode *c,*p;
bool red;
if(!n->left)
  {
  c=n->right;
  }
else if(!n->right)
  {
  c=n->left;
  }
else
  {
  RBTREE_TYPENode *o=n,*l;
  n=n->right;
  while((l=n->left))
    {
    n=l;
    }
  c=n->right;
  p=_RBTREE_TYPE_parent(n);
  red=_RBTREE_TYPE_red(n);
  if(c)
    {
    _RBTREE_TYPE_setParent(c,p);
    }
  if(p==o)
    {
    p->right=c;
    p=n;
    }
  else
    {
    p->left=c;
    }
  n->parent_red=o->parent_red;
  n->left=o->left;
  n->right=o->right;
  RBTREE_TYPENode *op;
  if((op=_RBTREE_TYPE_parent(o)))
    {
    if(op->left==o)
      {
      op->left=n;
      }
    else
      {
      op->right=n;
      }
    }
  else
    {
    t->root=n;
    }
  _RBTREE_TYPE_setParent(o->left,n);
  if(o->right)
    {
    _RBTREE_TYPE_setParent(o->right,n);
    }
  goto color;
  }
p=_RBTREE_TYPE_parent(n);
red=_RBTREE_TYPE_red(n);
if(c)
  {
  _RBTREE_TYPE_setParent(c,p);
  }
if(p)
  {
  if(p->left==n)
    {
    p->left=c;
    }
  else
    {
    p->right=c;
    }
  }
else
  {
  t->root=c;
  }
color:
if(!red)
  {
  while((!c||!_RBTREE_TYPE_red(c))&&(c!=t->root))
    {
    if(p->left==c)
      {
      RBTREE_TYPENode *o=p->right;
      if(_RBTREE_TYPE_red(o))
        {
        _RBTREE_TYPE_toBlack(o);
        _RBTREE_TYPE_toRed(p);
        _RBTREE_TYPE_rotateLeft(t,p);
        o=p->right;
        }
      if((!o->left||!_RBTREE_TYPE_red(o->left))&&
         (!o->right||!_RBTREE_TYPE_red(o->right)))
        {
        _RBTREE_TYPE_toRed(o);
        c=p;
        p=_RBTREE_TYPE_parent(c);
        }
      else
        {
        if(!o->right||!_RBTREE_TYPE_red(o->right))
          {
          RBTREE_TYPENode *oleft;
          if((oleft=o->left))
            {
            _RBTREE_TYPE_toBlack(oleft);
            }
          _RBTREE_TYPE_toRed(o);
          _RBTREE_TYPE_rotateRight(t,o);
          o=p->right;
          }
        _RBTREE_TYPE_setRed(o,_RBTREE_TYPE_red(p));
        _RBTREE_TYPE_toBlack(p);
        if(o->right)
          {
          _RBTREE_TYPE_toBlack(o->right);
          }
        _RBTREE_TYPE_rotateLeft(t,p);
        c=t->root;
        break;
        }
      }
    else
      {
      RBTREE_TYPENode *o=p->left;
      if(_RBTREE_TYPE_red(o))
        {
        _RBTREE_TYPE_toBlack(o);
        _RBTREE_TYPE_toRed(p);
        _RBTREE_TYPE_rotateRight(t,p);
        o=p->left;
        }
      if((!o->left||!_RBTREE_TYPE_red(o->left))&&
         (!o->right||!_RBTREE_TYPE_red(o->right)))
        {
        _RBTREE_TYPE_toRed(o);
        c=p;
        p=_RBTREE_TYPE_parent(c);
        }
      else
        {
        if(!o->left||!_RBTREE_TYPE_red(o->left))
          {
          RBTREE_TYPENode *oright;
          if((oright=o->right))
            {
            _RBTREE_TYPE_toBlack(oright);
            }
          _RBTREE_TYPE_toRed(o);
          _RBTREE_TYPE_rotateLeft(t,o);
          o=p->left;
          }
        _RBTREE_TYPE_setRed(o,_RBTREE_TYPE_red(p));
        _RBTREE_TYPE_toBlack(p);
        if(o->left)
          {
          _RBTREE_TYPE_toBlack(o->left);
          }
        _RBTREE_TYPE_rotateRight(t,p);
        c=t->root;
        break;
        }
      }
    }
  if(c)
    {
    _RBTREE_TYPE_toBlack(c);
    }
  }
Memory_free(destroyed);
--(t->size);
}

TRANSPROG_INLINE
void
RBTREE_TYPE_erase(RBTREE_TYPE *t,
                  RBTREE_TYPE_ElemType element)
{
RBTREE_TYPE_eraseNode(t,RBTREE_TYPE_find(t,element));
}

#endif // GUARD_MACRO

//----------------------------------------------------------------------------
'''

import sys
import os

if len(sys.argv)<4:
  sys.stderr.write(
             'usage: %s file_name element_type container_type [precode ...]\n'
             % sys.argv[0])
  sys.exit(1)

fileName=sys.argv[1]
elementType=sys.argv[2]
containerType=sys.argv[3]

guardMacro=os.path.normpath(fileName).upper().replace(os.path.sep,'_') \
                                             .replace(os.path.extsep,'_') \
                                             .replace(':','_')

thisFile='"%s"' % os.path.normpath(sys.argv[0]).replace(os.path.sep,'/')

specificCode='\n'.join(sys.argv[4:])
if specificCode:
  specificCode='\n// specific code\n'+specificCode

open(fileName,'w').write(content.replace('ELEM_TYPE',elementType)
                                .replace('RBTREE_TYPE',containerType)
                                .replace('GUARD_MACRO',guardMacro)
                                .replace('THIS_FILE',thisFile)
                                .replace('SPECIFIC_CODE',specificCode))

