#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

content='//'+'-'*76+r'''
// Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
// 
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef GUARD_MACRO
#define GUARD_MACRO 1
#line 44 THIS_FILE

#include "TransProg/memory.h"

SPECIFIC_CODE

typedef ELEM_TYPE VECTOR_TYPE_ElemType;

typedef int (*VECTOR_TYPE_CmpFnct)(const VECTOR_TYPE_ElemType *,
                                   const VECTOR_TYPE_ElemType *);

#if !defined VECTOR_TYPE_NO_DEFAULT_CMP
  TRANSPROG_INLINE
  int // -1  0  1
  VECTOR_TYPE_defaultCmpUpFnct(const VECTOR_TYPE_ElemType *a,
                               const VECTOR_TYPE_ElemType *b)
  {
  return *a==*b ? 0 : *a<*b ? -1 : 1;
  }

  TRANSPROG_INLINE
  int // -1  0  1
  VECTOR_TYPE_defaultCmpDownFnct(const VECTOR_TYPE_ElemType *a,
                                 const VECTOR_TYPE_ElemType *b)
  {
  return *a==*b ? 0 : *a<*b ? 1 : -1;
  }
#endif

typedef struct _VECTOR_TYPE
{
VECTOR_TYPE_ElemType *data;
intptr_t size;
intptr_t capacity;
} VECTOR_TYPE;

TRANSPROG_INLINE
void
VECTOR_TYPE_init(VECTOR_TYPE *v)
{
v->data=(VECTOR_TYPE_ElemType *)0;
v->size=0;
v->capacity=0;
}

TRANSPROG_INLINE
VECTOR_TYPE *
VECTOR_TYPE_new(void)
{
VECTOR_TYPE *v=Memory_alloc(1,VECTOR_TYPE);
VECTOR_TYPE_init(v);
return v;
}

TRANSPROG_INLINE
void
VECTOR_TYPE_destroy(VECTOR_TYPE *v)
{
if(v->data)
  {
  Memory_free(v->data);
  }
}

TRANSPROG_INLINE
void
VECTOR_TYPE_delete(VECTOR_TYPE *v)
{
VECTOR_TYPE_destroy(v);
Memory_free(v);
}

TRANSPROG_INLINE
void
VECTOR_TYPE_numaUsage(const VECTOR_TYPE *v,
                      intptr_t *numaUsage)
{
if(v->data)
  {
  Memory_numaUsage(v->data,numaUsage);
  }
}

TRANSPROG_INLINE
intptr_t
VECTOR_TYPE_capacity(const VECTOR_TYPE *v)
{
return v->capacity;
}

TRANSPROG_INLINE
void
VECTOR_TYPE_shrink(VECTOR_TYPE *v)
{
if(v->size)
  {
  if(v->size!=v->capacity)
    {
    v->data=Memory_reAlloc(v->data,v->size,VECTOR_TYPE_ElemType);
    v->capacity=v->size;
    }
  }
else if(v->data)
  {
  Memory_free(v->data);
  v->data=(VECTOR_TYPE_ElemType *)0;
  v->capacity=0;
  }
}

TRANSPROG_INLINE
void
VECTOR_TYPE_ensureMinCapacity(VECTOR_TYPE *v,
                              intptr_t capacity)
{
intptr_t newCapacity=v->capacity ? v->capacity : 16;
while(newCapacity<capacity)
  {
  newCapacity<<=1;
  }
if(newCapacity>v->capacity)
  {
  v->data=Memory_reAlloc(v->data,newCapacity,VECTOR_TYPE_ElemType);
  v->capacity=newCapacity;
  }
}

TRANSPROG_INLINE
intptr_t
VECTOR_TYPE_size(const VECTOR_TYPE *v)
{
return v->size;
}

TRANSPROG_INLINE
void
VECTOR_TYPE_clear(VECTOR_TYPE *v)
{
v->size=0;
}

TRANSPROG_INLINE
void
VECTOR_TYPE_setSize(VECTOR_TYPE *v,
                    intptr_t newSize)
{
if(newSize>v->capacity)
  {
  VECTOR_TYPE_ensureMinCapacity(v,newSize);
  }
v->size=newSize;
}

TRANSPROG_INLINE
VECTOR_TYPE_ElemType * // element array
VECTOR_TYPE_stealData(VECTOR_TYPE *v)
{
VECTOR_TYPE_ElemType *data=v->data;
v->data=(VECTOR_TYPE_ElemType *)0;
v->size=0;
v->capacity=0;
return data;
}

TRANSPROG_INLINE
VECTOR_TYPE_ElemType * // element array
VECTOR_TYPE_data(const VECTOR_TYPE *v)
{
return v->data;
}

TRANSPROG_INLINE
VECTOR_TYPE_ElemType // index'th element
VECTOR_TYPE_get(const VECTOR_TYPE *v,
                intptr_t index)
{
return v->data[index];
}

TRANSPROG_INLINE
VECTOR_TYPE_ElemType // first element
VECTOR_TYPE_first(const VECTOR_TYPE *v)
{
return v->data[0];
}

TRANSPROG_INLINE
VECTOR_TYPE_ElemType // last element
VECTOR_TYPE_last(const VECTOR_TYPE *v)
{
return v->data[v->size-1];
}

TRANSPROG_INLINE
void
VECTOR_TYPE_set(VECTOR_TYPE *v,
                intptr_t index,
                VECTOR_TYPE_ElemType element)
{
v->data[index]=element;
}

TRANSPROG_INLINE
void
VECTOR_TYPE_push(VECTOR_TYPE *v,
                 VECTOR_TYPE_ElemType element)
{
if(v->size==v->capacity)
  {
  VECTOR_TYPE_ensureMinCapacity(v,v->size+1);
  }
v->data[v->size++]=element;
}

TRANSPROG_INLINE
void
VECTOR_TYPE_pushMany(VECTOR_TYPE *v,
                     const VECTOR_TYPE_ElemType *elements,
                     intptr_t elementCount)
{
intptr_t newSize=v->size+elementCount;
if(newSize>v->capacity)
  {
  VECTOR_TYPE_ensureMinCapacity(v,newSize);
  }
memcpy(v->data+v->size,elements,elementCount*sizeof(VECTOR_TYPE_ElemType));
v->size=newSize;
}

TRANSPROG_INLINE
VECTOR_TYPE_ElemType
VECTOR_TYPE_pop(VECTOR_TYPE *v)
{
return v->data[--(v->size)];
}

TRANSPROG_INLINE
void
VECTOR_TYPE_popMany(VECTOR_TYPE *v,
                    intptr_t elementCount)
{
v->size-=elementCount;
}

TRANSPROG_INLINE
void
VECTOR_TYPE_insert(VECTOR_TYPE *v,
                   intptr_t index,
                   VECTOR_TYPE_ElemType element)
{
if(v->size==v->capacity)
  {
  VECTOR_TYPE_ensureMinCapacity(v,v->size+1);
  }
VECTOR_TYPE_ElemType *p=v->data+index;
memmove(p+1,p,((v->size++)-index)*sizeof(VECTOR_TYPE_ElemType));
*p=element;
}

TRANSPROG_INLINE
void
VECTOR_TYPE_insertMany(VECTOR_TYPE *v,
                       intptr_t index,
                       const VECTOR_TYPE_ElemType *elements,
                       intptr_t elementCount)
{
intptr_t newSize=v->size+elementCount;
if(newSize>v->capacity)
  {
  VECTOR_TYPE_ensureMinCapacity(v,newSize);
  }
VECTOR_TYPE_ElemType *p=v->data+index;
memmove(p+elementCount,p,(v->size-index)*sizeof(VECTOR_TYPE_ElemType));
memcpy(p,elements,elementCount*sizeof(VECTOR_TYPE_ElemType));
v->size=newSize;
}

TRANSPROG_INLINE
void
VECTOR_TYPE_fastErase(VECTOR_TYPE *v,
                      intptr_t index)
{
v->data[index]=v->data[--(v->size)];
}

TRANSPROG_INLINE
void
VECTOR_TYPE_erase(VECTOR_TYPE *v,
                  intptr_t index)
{
VECTOR_TYPE_ElemType *p=v->data+index;
memmove(p,p+1,((--(v->size))-index)*sizeof(VECTOR_TYPE_ElemType));
}

TRANSPROG_INLINE
void
VECTOR_TYPE_eraseMany(VECTOR_TYPE *v,
                      intptr_t index,
                      intptr_t elementCount)
{
VECTOR_TYPE_ElemType *p=v->data+index;
v->size-=elementCount;
memmove(p,p+elementCount,(v->size-index)*sizeof(VECTOR_TYPE_ElemType));
}

TRANSPROG_INLINE
void
VECTOR_TYPE_sort(VECTOR_TYPE *v,
                 VECTOR_TYPE_CmpFnct cmp)
{
#if !defined VECTOR_TYPE_NO_DEFAULT_CMP
  if(!cmp)
    {
    cmp=VECTOR_TYPE_defaultCmpUpFnct;
    }
#endif
qsort(v->data,v->size,sizeof(VECTOR_TYPE_ElemType),
      (int (*)(const void *,const void *))cmp);
}

TRANSPROG_INLINE
intptr_t // insertion index
VECTOR_TYPE_sortedInsertionIndex(const VECTOR_TYPE *v,
                                 VECTOR_TYPE_CmpFnct cmp,
                                 VECTOR_TYPE_ElemType element)
{
#if !defined VECTOR_TYPE_NO_DEFAULT_CMP
  if(!cmp)
    {
    cmp=VECTOR_TYPE_defaultCmpUpFnct;
    }
#endif
intptr_t start=0;
intptr_t end=v->size;
while(start!=end)
  {
  intptr_t mid=(start+end)>>1;
  if((*cmp)(&element,&v->data[mid])<0)
    {
    end=mid;
    }
  else
    {
    start=mid+1;
    }
  }
return start;
}

TRANSPROG_INLINE
intptr_t // insertion index
VECTOR_TYPE_sortedInsert(VECTOR_TYPE *v,
                         VECTOR_TYPE_CmpFnct cmp,
                         VECTOR_TYPE_ElemType element)
{
intptr_t index=VECTOR_TYPE_sortedInsertionIndex(v,cmp,element);
VECTOR_TYPE_insert(v,index,element);
return index;
}

TRANSPROG_INLINE
intptr_t // found index or -1
VECTOR_TYPE_sortedFind(const VECTOR_TYPE *v,
                       VECTOR_TYPE_CmpFnct cmp,
                       VECTOR_TYPE_ElemType element)
{
#if !defined VECTOR_TYPE_NO_DEFAULT_CMP
  if(!cmp)
    {
    cmp=VECTOR_TYPE_defaultCmpUpFnct;
    }
#endif
intptr_t start=0;
intptr_t end=v->size;
while(start!=end)
  {
  intptr_t mid=(start+end)>>1;
  int r=(*cmp)(&element,&v->data[mid]);
  if(!r) return mid;
  if(r<0)
    {
    end=mid;
    }
  else
    {
    start=mid+1;
    }
  }
return -1;
}

TRANSPROG_INLINE
intptr_t // found index or -1
VECTOR_TYPE_find(const VECTOR_TYPE *v,
                 VECTOR_TYPE_CmpFnct cmp,
                 VECTOR_TYPE_ElemType element)
{
if(cmp)
  {
  for(intptr_t index=0,index_end=v->size;index<index_end;++index)
    {
    if(!(*cmp)(&element,&v->data[index]))
      {
      return index;
      }
    }
  }
#if !defined VECTOR_TYPE_NO_DEFAULT_CMP
  else
    {
    for(intptr_t index=0,index_end=v->size;index<index_end;++index)
      {
      if(element==v->data[index])
        {
        return index;
        }
      }
    }
#endif
return -1;
}

#endif // GUARD_MACRO

//----------------------------------------------------------------------------
'''

import sys
import os

if len(sys.argv)<4:
  sys.stderr.write(
             'usage: %s file_name element_type container_type [precode ...]\n'
             % sys.argv[0])
  sys.exit(1)

fileName=sys.argv[1]
elementType=sys.argv[2]
containerType=sys.argv[3]

guardMacro=os.path.normpath(fileName).upper().replace(os.path.sep,'_') \
                                             .replace(os.path.extsep,'_') \
                                             .replace(':','_')

thisFile='"%s"' % os.path.normpath(sys.argv[0]).replace(os.path.sep,'/')

specificCode='\n'.join(sys.argv[4:])
if specificCode:
  specificCode='\n// specific code\n'+specificCode

open(fileName,'w').write(content.replace('ELEM_TYPE',elementType)
                                .replace('VECTOR_TYPE',containerType)
                                .replace('GUARD_MACRO',guardMacro)
                                .replace('THIS_FILE',thisFile)
                                .replace('SPECIFIC_CODE',specificCode))

