//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/text.h"

static
intptr_t // inserted chars
Text_appendArgs(Text *t,
                const char *format,
                va_list args)
{
intptr_t tLen=Text_length(t);
intptr_t fLen=strlen(format);
Text_ensureMinCapacity(t,tLen+fLen);
for(;;)
  {
  char *str=Text_get(t)+tLen;
  intptr_t remaining=t->v.capacity-tLen;
#if defined _WIN32
  intptr_t r=vsnprintf(str,remaining,format,args);
#else
  va_list tmpArgs;
  va_copy(tmpArgs,args);
  intptr_t r=vsnprintf(str,remaining,format,tmpArgs);
  va_end(tmpArgs);
#endif
  if((r>=0)&&(r<remaining))
    {
    t->v.size+=r;
    return r;
    }
  else
    {
    Text_ensureMinCapacity(t,t->v.capacity+fLen);
    }
  }
}

intptr_t // inserted chars
Text_insert(Text *t,
            intptr_t index,
            const char *format,
            ...)
{
Text t2;
Text_init(&t2);
va_list args;
va_start(args,format);
intptr_t r=Text_appendArgs(&t2,format,args);
va_end(args);
Text_insertText(t,index,&t2);
Text_destroy(&t2);
return r;
}

intptr_t // inserted chars
Text_append(Text *t,
            const char *format,
            ...)
{
va_list args;
va_start(args,format);
intptr_t r=Text_appendArgs(t,format,args);
va_end(args);
return r;
}

//----------------------------------------------------------------------------
